# -*- coding: utf-8 -*-
#
# The data engine that provides DBus service for handling a single instance of a progress group.
#
# Created for using with Make Progress! plasmoid but has a general purpose.
# The plasmoid or another program that uses this engine will be refered later as "the program".
#
# >>> Initialization: <<<
#
# The program should register its instance name via service for 'instance name' source.
# The engine will search through occupied names and could modify the given name to create a valid available name.
# Then it exports the DBus root object to provide interface for modification from outside.
# DBus service name will be concatenated global var bus_name + '.make-progress.' + instance name.
# After that engine updates source data with the instance name and the program should react properly on this event.
#
# >>> Creating a progress: <<<
#
# If instance name is registered then the engine will provide a new data source 'progress factory'.
# A service for this source could be used to register a new progress that was created via program interface.
# The program should provide an id of progress item that is unique among all items created in the program.
# This id is used for futher intercommunication and for separating replies from engine.
# If progress is created externally via DBus interface then engine creates id by itself
# and updates the data source to advertise a new progress to the program.
# This id is guaranteed to be unique within a set of items created from outside.
#
# When progress had been created the engine exports the DBus object /ProgressName,
# where ProgressName is a generated name based on the id and the source of creation.
# The 'progress factory' is updated then. It contains basic information about all existing items.
# Also engine provides a new source 'progress: ProgressName' that allows program to send and receive progress updates.
#
# >>> Modifying a progress: <<<
#
# Properties of a progress could be changed very frequently by both a user and the program.
# For example, the progress could be a timer and user want to correct it from time to time.
# So we need a protocol here to keep progress properties synchronized.
# The higher priority have values that are set via DBus because it's values from user or user's script.
# User could also correct values via interface, but doing this simultaneously with his script has no sense.
#
# The program uses service for the source 'progress: ProgressName' to update the data.
# In the regular mode the engine simply updates it's internal data so user can get it through DBus.
# If parameter was modified externally then the engine saves the value and switches to the request mode.
# Then it updates the source and shows to the program that some parameters were set by user.
# The program modifies parameter of the progress and tells the engine by service that the thing is updated.
# If the engine receives the same data as it currently has then it switches to the regular mode.
#
# >>> Removing a progress: <<<
#
# If we simply delete a progress from the engine then the program will not see the progress info
# in the data of the source 'progress factory' but the program can't delete that item
# because it can be the case when user created a new progress with interface and it hasn't appeared in the source yet.
# If we simply delete a progress from the program and send info about it by service
# then for some time the engine will show that progress in the source 'progress factory'
# and the program would think that it is a new progress and would create it.
# So we need a protocol here to keep progress state synchronized.
#
# When the engine decides to remove a progress it marks the progress as removed, removes DBus iface and updates the source.
# The program receives this data, actually removes the item and says to the engine that it's ok to delete this item.
# Then the engine actually removes the progress object here.
#
# When the program decides to remove a progress it sends a request for that to the engine.
# The engine receives this data and this state is equal to the case when the engine itself wants to delete the item.
#
# >>> Notes: <<<
#
# * each source name is prefixed with a special tag, see MPSource().
# * to enable or disable debug output use 'debug' global flag
#
# @license GPL 2 (http://www.gnu.org/licenses/gpl.html)
# @author Nikita Melnichenko [http://nikita.melnichenko.name]
# @copyright Copyright 2010, Nikita Melnichenko


import os.path
import dbus
import dbus.service
from dbus.mainloop.qt import DBusQtMainLoop
import string

from PyQt4.QtCore import *
from PyKDE4.plasma import Plasma
from PyKDE4.plasmascript import DataEngine



# a kind of global settings

# bus name of DBus service
bus_name = 'name.melnichenko.nikita'
# flag: enable debug
debug = False

# debug print
def debug_print(text):
	if debug:
		print text

# prefix all sources of engine
# used in JS plasmoid with API v1 when all engines are attached to single slot
# this helps to separate the engine replies from other replies
def MPSource(name):
	return '[jdbus] ' + name

# create item name for internally created items
def MPInternalItemName(id):
	return 'i' + str(id)

# create item name for externally created items
def MPExternalItemName(id):
	return 'e' + str(id)



# a representaion of a basic progress in the DBus service
class DBusProgress(dbus.service.Object):
	def __init__(self, instance, id, name, origin, type):
		debug_print('>>> MakeProgress, DBusProgress: constructor(' + str(id) + ', ' + name + ', ' + str(origin) + ', ' + type + ')')
		dbus.service.Object.__init__(self, dbus.SessionBus(), '/' + name)
		self.instance = instance
		self.id = id
		self.name = name
		self.origin = origin
		self.type = type
		self.minimum = 0
		self.maximum = 100
		self.value = 0

		self.is_minimum_requested = False
		self.is_maximum_requested = False
		self.is_value_requested = False
		self.is_removed = False

		# update engine data
		self.instance.engine.setData(MPSource('progress factory'), self.name,
			{ "id" : self.id, "origin" : self.origin,  "type": self.type })
		self.UpdateEngineData("id", self.id)
		self.UpdateEngineData("name", self.name)
		self.UpdateEngineData("origin", self.origin)
		self.UpdateEngineData("type", self.type)
		self.UpdateEngineData("minimum", self.minimum)
		self.UpdateEngineData("maximum", self.maximum)
		self.UpdateEngineData("value", self.value)

	def __del__(self):
		debug_print('>>> MakeProgress, DBusProgress[' + self.name + ']: destructor')

	def UpdateEngineData(self, key, value):
		self.instance.engine.setData(MPSource('progress: ' + self.name), key, value)

	def RemoveEngineData(self, key):
		self.instance.engine.removeData(MPSource('progress: ' + self.name), key)

	@dbus.service.method(bus_name, in_signature='', out_signature='s')
	def Type(self):
		debug_print('>>> MakeProgress, DBusProgress[' + self.name + ']: Type')
		return self.type

	@dbus.service.method(bus_name, in_signature='', out_signature='d')
	def GetMinimum(self):
		debug_print('>>> MakeProgress, DBusProgress[' + self.name + ']: GetMinimum')
		return self.minimum

	@dbus.service.method(bus_name, in_signature='d', out_signature='')
	def SetMinimum(self, value):
		self.SetProperty('minimum', float(value))

	@dbus.service.method(bus_name, in_signature='', out_signature='d')
	def GetMaximum(self):
		debug_print('>>> MakeProgress, DBusProgress[' + self.name + ']: GetMaximum')
		return self.maximum

	@dbus.service.method(bus_name, in_signature='d', out_signature='')
	def SetMaximum(self, value):
		self.SetProperty('maximum', float(value))

	@dbus.service.method(bus_name, in_signature='', out_signature='d')
	def GetValue(self):
		debug_print('>>> MakeProgress, DBusProgress[' + self.name + ']: GetValue')
		return self.value

	@dbus.service.method(bus_name, in_signature='d', out_signature='')
	def SetValue(self, value):
		self.SetProperty('value', float(value))

	def SetProperty(self, prop, value):
		debug_print('>>> MakeProgress, DBusProgress[' + self.name + ']: SetProperty(' + prop + ', ' + str(value) + ')')
		self.__dict__[prop] = value
		self.UpdateEngineData(prop, value)
		# switch to request mode
		self.__dict__['is_' + prop + '_requested'] = True
		self.UpdateEngineData("request: " + prop, True)

	def SetPropertyFromJob(self, prop, is_requested, value):
		if self.is_removed or prop not in self.__dict__:
			return False
		debug_print('>>> MakeProgress, DBusProgress[' + self.name + ']: SetPropertyFromJob(' + prop + ', ' + str(is_requested) + ', ' + str(value) + ')')
		prop_value = self.__dict__[prop]
		is_prop_requested = self.__dict__['is_' + prop + '_requested']
		if is_prop_requested:
			if is_requested:
				if prop_value == value:
					debug_print('>>> MakeProgress, DBusProgress[' + self.name + '], SetPropertyFromJob: request reply')
					# switch to normal mode
					self.__dict__['is_' + prop + '_requested'] = False
					self.RemoveEngineData("request: " + prop)
					return True
				else:
					debug_print('>>> MakeProgress, DBusProgress[' + self.name + '], SetPropertyFromJob: requested value not matched')
					return False
			else:
				debug_print('>>> MakeProgress, DBusProgress[' + self.name + '], SetPropertyFromJob: non-requested value ignored')
				return False
		else:
			debug_print('>>> MakeProgress, DBusProgress[' + self.name + '], SetPropertyFromJob: value is set')
			self.__dict__[prop] = value
			self.UpdateEngineData(prop, self.__dict__[prop])
			return True

	def ScheduleRemoval(self):
		debug_print('>>> MakeProgress, DBusProgress[' + self.name + ']: ScheduleRemoval')
		self.remove_from_connection()
		self.instance.engine.setData(MPSource('progress factory'), self.name,
			{ "id" : self.id, "origin" : self.origin, "type": self.type, "request: remove": True })
		self.is_removed = True



# a representaion of a basic progress in the DBus service
class DBusLiveProgress(DBusProgress):
	def __init__(self, instance, id, name, origin, type):
		debug_print('>>> MakeProgress, DBusLiveProgress: constructor(' + str(id) + ', ' + name + ', ' + str(origin) + ', ' + type + ')')
		DBusProgress.__init__(self, instance, id, name, origin, type)
		self.active = False
		self.speed = 0

		self.is_active_requested = False
		self.is_speed_requested = False

		self.UpdateEngineData("active", self.active)
		self.UpdateEngineData("speed", self.speed)

	@dbus.service.method(bus_name, in_signature='', out_signature='')
	def Start(self):
		debug_print('>>> MakeProgress, DBusLiveProgress[' + self.name + ']: Start')
		self.SetProperty('active', True)

	@dbus.service.method(bus_name, in_signature='', out_signature='')
	def Stop(self):
		debug_print('>>> MakeProgress, DBusLiveProgress[' + self.name + ']: Stop')
		self.SetProperty('active', False)

	@dbus.service.method(bus_name, in_signature='', out_signature='b')
	def IsStarted(self):
		debug_print('>>> MakeProgress, DBusLiveProgress[' + self.name + ']: IsStarted')
		return self.active

	@dbus.service.method(bus_name, in_signature='', out_signature='d')
	def GetSpeed(self):
		debug_print('>>> MakeProgress, DBusProgress[' + self.name + ']: GetSpeed')
		return self.speed

	@dbus.service.method(bus_name, in_signature='d', out_signature='')
	def SetSpeed(self, value):
		self.SetProperty('speed', float(value))



# instance of the program
class DBusInstance(dbus.service.Object):
	def __init__(self, engine):
		debug_print('>>> MakeProgress, DBusInstance: constructor')
		dbus.service.Object.__init__(self, dbus.SessionBus(), '/')
		self.engine = engine
		self.items = []
		# the next unique id of external item
		self.next_id = 1

	def __del__(self):
		debug_print('>>> MakeProgress, DBusInstance: destructor')

	@dbus.service.method(bus_name, in_signature='', out_signature='s')
	def Version(self):
		debug_print('>>> MakeProgress, DBusInstance: Version')
		return str(self.engine.package().metadata().version())

	@dbus.service.method(bus_name, in_signature='', out_signature='s')
	def Author(self):
		debug_print('>>> MakeProgress, DBusInstance: Author')
		return str(self.engine.package().metadata().author())

	@dbus.service.method(bus_name, in_signature='', out_signature='s')
	def Homepage(self):
		debug_print('>>> MakeProgress, DBusInstance: Homepage')
		return str(self.engine.package().metadata().website())

	def ItemByName(self, name):
		for i in self.items:
			if i.name == name:
				return i
		return None

	@dbus.service.method(bus_name, in_signature='', out_signature='as')
	def Items(self):
		debug_print('>>> MakeProgress, DBusInstance: Items')
		names = []
		for i in self.items:
			names.append(i.name)
		return names

	def CreateItem(self, id, name, origin, type):
		if name in self.Items():
			debug_print('>>> MakeProgress, DBusInstance, CreateItem: name ' + name + ' already exists!')
			return False
		if type == 'Live' or type == 'Timer':
			item = DBusLiveProgress(self, id, name, origin, type)
		else:
			item = DBusProgress(self, id, name, origin, type)
		self.items.append(item)
		return True


	def AddInternalItem(self, id, type):
		debug_print('>>> MakeProgress, DBusInstance: AddInternalItem(' + str(id) + ', ' + type + ')')
		name = MPInternalItemName(id)
		return self.CreateItem(id, name, 'Internal', type)

	@dbus.service.method(bus_name, in_signature='s', out_signature='s')
	def AddItem(self, type):
		debug_print('>>> MakeProgress, DBusInstance: AddItem')
		id = self.next_id
		name = MPExternalItemName(id)
		self.next_id += 1
		self.CreateItem(id, name, 'External', str(type))
		return name

	# this method is called RemoveItem (for exporting to DBus) but actually it only schedules for removal
	# the method that removes item is called DestroyItem
	@dbus.service.method(bus_name, in_signature='s', out_signature='b')
	def RemoveItem(self, name):
		debug_print('>>> MakeProgress, DBusInstance: RemoveItem(' + name + ')')
		item = self.ItemByName(str(name))
		if item == None:
			return False
		item.ScheduleRemoval()
		return True

	def DestroyItem(self, name):
		debug_print('>>> MakeProgress, DBusInstance: DestroyItem(' + name + ')')
		self.items = filter(lambda i: i.name != name, self.items)
		self.engine.deleteSource(MPSource('progress: ' + name))
		return True



# DBus service for this data engine; should be a singleton
class MakeProgressDBusService():
	def __init__(self, engine):
		debug_print('>>> MakeProgressDBusService: constructor')
		# create a service
		self.dbus_service_bus_name = dbus.service.BusName(MakeProgressEngine.ComposeBusName(engine.instance_name), dbus.SessionBus())
		# export the instance object
		self.instance = DBusInstance(engine)

	def __del__(self):
		debug_print('>>> MakeProgressDBusService: destructor')
		self.instance.remove_from_connection()



class JobInstanceName(Plasma.ServiceJob):
	def __init__(self, source, operation, parameters, parent = None):
		debug_print('>>> MakeProgress, JobInstanceName: constructor')
		Plasma.ServiceJob.__init__(self, source.objectName(), operation, parameters, parent)
		self.source = source

	def __del__(self):
		debug_print('>>> MakeProgress, JobInstanceName: destructor')

	def start(self):
		debug_print('>>> MakeProgress, JobInstanceName: start(' + self.operationName() + ')')
		if self.operationName() == 'SetName':
			if QString('name') in self.parameters():
				name = str(self.parameters()[QString('name')].toString().toUtf8())
				debug_print('>>> MakeProgress, JobInstanceName, SetName: name = ' + name)
				self.source.SetInstanceName(name)
				self.setResult(True)
				return
			else:
				debug_print('>>> MakeProgress, JobInstanceName, SetName: wrong parameters!')
				self.setResult(False)
				return
		debug_print('>>> MakeProgress, JobInstanceName: unhandled!')
		self.setResult(False)



class ServiceInstanceName(Plasma.Service):
	def __init__(self, engine):
		debug_print('>>> MakeProgress, ServiceInstanceName: constructor')
		Plasma.Service.__init__(self, engine)
		self.engine = engine
		self.setName('instance name')

	def __del__(self):
		debug_print('>>> MakeProgress, ServiceInstanceName: destructor')

	def createJob(self, operation, parameters):
		debug_print('>>> MakeProgress, ServiceInstanceName: createJob')
		return JobInstanceName(self.parent(), operation, parameters, self)

	def registerOperationsScheme(self):
		debug_print('>>> MakeProgress, ServiceInstanceName: registerOperationsScheme')
		path = self.engine.getOperationSchemePath('InstanceName.xml')
		if path.isEmpty():
			print '>>> MakeProgress, ServiceInstanceName: no scheme found'
			return
		self.setOperationsScheme(QFile(path))



class JobProgressFactory(Plasma.ServiceJob):
	def __init__(self, source, operation, parameters, parent = None):
		debug_print('>>> MakeProgress, JobProgressFactory: constructor')
		Plasma.ServiceJob.__init__(self, source.objectName(), operation, parameters, parent)
		self.source = source
		self.instance = source.dbus_service.instance # existance has been check before creating a service

	def __del__(self):
		debug_print('>>> MakeProgress, JobProgressFactory: destructor')

	def start(self):
		debug_print('>>> MakeProgress, JobProgressFactory: start(' + self.operationName() + ')')
		if self.operationName() == 'RegisterItem':
			if QString('id') in self.parameters() and QString('type') in self.parameters():
				id = self.parameters()[QString('id')].toInt()[0]
				type = str(self.parameters()[QString('type')].toString())
				debug_print('>>> MakeProgress, JobProgressFactory, RegisterItem: id = ' + str(id) + ', type = ' + type)
				self.setResult(self.instance.AddInternalItem(id, type))
				return
			else:
				debug_print('>>> MakeProgress, JobProgressFactory, RegisterItem: wrong parameters!')
				self.setResult(False)
				return
		if self.operationName() == 'ScheduleItemRemoval':
			if QString('name') in self.parameters():
				name = str(self.parameters()[QString('name')].toString())
				debug_print('>>> MakeProgress, JobProgressFactory, ScheduleItemRemoval: name = ' + name)
				self.setResult(self.instance.RemoveItem(name))
				return
			else:
				debug_print('>>> MakeProgress, JobProgressFactory, ScheduleItemRemoval: wrong parameters!')
				self.setResult(False)
				return
		if self.operationName() == 'RemoveItem':
			if QString('name') in self.parameters():
				name = str(self.parameters()[QString('name')].toString())
				debug_print('>>> MakeProgress, JobProgressFactory, RemoveItem: name = ' + name)
				self.setResult(self.instance.DestroyItem(name))
				return
			else:
				debug_print('>>> MakeProgress, JobProgressFactory, RemoveItem: wrong parameters!')
				self.setResult(False)
				return
		debug_print('>>> MakeProgress, JobProgressFactory: unhandled!')
		self.setResult(False)



class ServiceProgressFactory(Plasma.Service):
	def __init__(self, engine):
		debug_print('>>> MakeProgress, ServiceProgressFactory: constructor')
		Plasma.Service.__init__(self, engine)
		self.engine = engine
		self.setName('progress factory')

	def __del__(self):
		debug_print('>>> MakeProgress, ServiceProgressFactory: destructor')

	def createJob(self, operation, parameters):
		debug_print('>>> MakeProgress, ServiceProgressFactory: createJob')
		return JobProgressFactory(self.parent(), operation, parameters, self)

	def registerOperationsScheme(self):
		debug_print('>>> MakeProgress, ServiceProgressFactory: registerOperationsScheme')
		path = self.engine.getOperationSchemePath('ProgressFactory.xml')
		if path.isEmpty():
			print '>>> MakeProgress, ServiceProgressFactory: no scheme found'
			return
		self.setOperationsScheme(QFile(path))



class JobProgressProps(Plasma.ServiceJob):
	def __init__(self, object, source, operation, parameters, parent = None):
		debug_print('>>> MakeProgress, JobProgressProps: constructor')
		Plasma.ServiceJob.__init__(self, source.objectName(), operation, parameters, parent)
		self.source = source
		self.object = object

	def __del__(self):
		debug_print('>>> MakeProgress, JobProgressProps: destructor')

	def start(self):
		debug_print('>>> MakeProgress, JobProgressProps: start(' + self.operationName() + ')')
		if self.operationName() == 'SetValue' \
		or self.operationName() == 'SetMinimum' \
		or self.operationName() == 'SetMaximum' \
		or self.operationName() == 'SetSpeed':
			# the reason why parameters are prefixed with operationName is simple: kcfg format states that entry name is a key
			# prooflink: http://developer.kde.org/documentation/tutorials/kconfigxt/kconfigxt.html
			# ...
			# 4.3 The individual entries must have at least a name or a key.
			# The name is used to create accessor and modifier functions. It's also used as the key in the config file.
			# ...
			# So groups don't help here while we have two types of values: Double and Bool
			if QString(self.operationName() + '_is_requested') in self.parameters() \
			and QString(self.operationName() + '_value') in self.parameters() :
				is_requested = self.parameters()[QString(self.operationName() + '_is_requested')].toBool()
				value = self.parameters()[QString(self.operationName() + '_value')].toDouble()[0]
				debug_print('>>> MakeProgress, JobProgressProps, ' + self.operationName() + ': is_requested = ' + str(is_requested) + ', value = ' + str(value))
				if self.operationName() == 'SetValue':
					prop = 'value'
				if self.operationName() == 'SetMinimum':
					prop = 'minimum'
				if self.operationName() == 'SetMaximum':
					prop = 'maximum'
				if self.operationName() == 'SetSpeed':
					prop = 'speed'
				self.setResult(self.object.SetPropertyFromJob(prop, is_requested, value))
				return
			else:
				debug_print('>>> MakeProgress, JobProgressProps, ' + self.operationName() + ': wrong parameters!')
				self.setResult(False)
				return
		if self.operationName() == 'SetActive':
			if QString('SetActive_is_requested') in self.parameters() and QString('SetActive_value') in self.parameters() :
				is_requested = self.parameters()[QString('SetActive_is_requested')].toBool()
				value = self.parameters()[QString('SetActive_value')].toBool()
				debug_print('>>> MakeProgress, JobProgressProps, SetActive: is_requested = ' + str(is_requested) + ', value = ' + str(value))
				self.setResult(self.object.SetPropertyFromJob('active', is_requested, value))
				return
			else:
				debug_print('>>> MakeProgress, JobProgressProps, SetActive: wrong parameters!')
				self.setResult(False)
				return
		debug_print('>>> MakeProgress, JobProgressProps: unhandled!')
		self.setResult(False)



class ServiceProgressProps(Plasma.Service):
	def __init__(self, object, engine):
		debug_print('>>> MakeProgress, ServiceProgressProps: constructor')
		Plasma.Service.__init__(self, engine)
		self.engine = engine
		self.setName('progress props')
		self.object = object

	def __del__(self):
		debug_print('>>> MakeProgress, ServiceProgressProps: destructor')

	def createJob(self, operation, parameters):
		debug_print('>>> MakeProgress, ServiceProgressProps: createJob')
		return JobProgressProps(self.object, self.parent(), operation, parameters, self)

	def registerOperationsScheme(self):
		debug_print('>>> MakeProgress, ServiceProgressProps: registerOperationsScheme')
		path = self.engine.getOperationSchemePath('ProgressProps.xml')
		if path.isEmpty():
			print '>>> MakeProgress, ServiceProgressProps: no scheme found'
			return
		self.setOperationsScheme(QFile(path))



class MakeProgressEngine(DataEngine):
	def __init__(self, parent, args = None):
		debug_print('>>> MakeProgressEngine: constructor')
		DataEngine.__init__(self, parent)
		# active source names (initial sources are added in the init function)
		self.active_sources = []
		# instance name needs to be registered
		self.instance_name = ''
		# init the default mainloop (needs to be done before the first DBus call)
		dbus.mainloop.qt.DBusQtMainLoop(set_as_default=True)
		# no dbus service until instance name is registered
		self.dbus_service = None
		# all services
		self.services = {}

	def __del__(self):
		debug_print('>>> MakeProgressEngine: destructor')

	def init(self):
		debug_print('>>> MakeProgressEngine: init')
		# update immediately on every update request
		self.setMinimumPollingInterval(0)
		# create initial sources
		sources = [ MPSource('version'), MPSource('instance name') ]
		for i in sources:
			self.createSource(i)
		# set version
		self.setData(MPSource('version'), 'version', self.package().metadata().version())

	def createSource(self, source):
		debug_print('>>> MakeProgressEngine: createSource(' + source + ')')
		self.active_sources.append(source)
		container = Plasma.DataContainer(self)
		container.setObjectName(source)
		# Delete a fake data key. This will trigger the update flag inside container.
		# Without it JS plasmoid would not handle update requests.
		# Inconvenience is that dataUpdated will be called with empty data once
		# Tested on 4.4.1; we can remove this hack later if that behavior will be fixed
		container.setData('fake', QVariant())
		self.addSource(container)

	def deleteSource(self, source):
		debug_print('>>> MakeProgressEngine: deleteSource(' + source + ')')
		self.removeSource(source)
		if source in self.services:
			del self.services[source]

	def updateSourceEvent(self, source):
		debug_print('>>> MakeProgressEngine: updateSourceEvent(' + source + ')')
		return True

	def sourceRequestEvent(self, source):
		debug_print('>>> MakeProgressEngine: sourceRequestEvent(' + source + ')')
		if source not in self.active_sources:
			return False
		return True

	def sources(self):
		debug_print('>>> MakeProgressEngine: sources')
		return self.active_sources
	
	def getOperationSchemePath(self, xml_file_name):
		# filePath doesn't work in KDE SC 4.6.2
		# path = self.package().filePath('config', xml_file_name)
		path = os.path.join(str(self.package().path()), 'contents', 'config', xml_file_name)
		debug_print('>>> MakeProgressEngine: getOperationSchemePath(' + xml_file_name + ') = ' + path)
		return QString(path)

	def serviceForSource(self, source):
		source = str(source) # convert source to the native type or self.services will freak out
		debug_print('>>> MakeProgressEngine: serviceForSource(' + source + ')')
		if source in self.services:
			debug_print('>>> MakeProgressEngine: service already created')
			return self.services[source]
		debug_print('>>> MakeProgressEngine: service is not created yet')
		if source == MPSource('instance name'):
			self.services[source] = ServiceInstanceName(self)
			return self.services[source]
		if self.dbus_service:
			if source == MPSource('progress factory'):
				self.services[source] = ServiceProgressFactory(self)
				return self.services[source]
			for i in self.dbus_service.instance.items:
				if source == MPSource('progress: ' + i.name):
					self.services[source] = ServiceProgressProps(i, self)
					return self.services[source]
		debug_print('>>> MakeProgressEngine: service name is not valid')
		return None

	def SetInstanceName(self, name):
		debug_print('>>> MakeProgressEngine: SetInstanceName(' + name + ')')
		# find a good name (here are first DBus calls)
		self.instance_name = MakeProgressEngine.FindAvailableInstanceName(name)
		debug_print('>>> MakeProgressEngine: SetInstanceName, instance name = ' + self.instance_name)
		# create the DBus service
		self.dbus_service = MakeProgressDBusService(self)
		# advertise
		self.setData(MPSource('instance name'), 'name', self.instance_name)
		self.createSource(MPSource('progress factory'))
		return True

	@staticmethod
	def FindAvailableInstanceName(input_name):
		# compose base name from input name (remove all chars that aren't ascii and digits)
		basename = filter(lambda x: x.isdigit() or x in string.ascii_letters, input_name)
		# dbus doesn't allow digit after dot so we prefix it
		if not basename or basename[0].isdigit():
			basename = '_' + basename
		debug_print('>>> MakeProgressEngine: SetInstanceName, base for instance name = ' + basename)
		# search between basename, basename2, basename3, ...
		i = 1
		name = basename
		try:
			object = dbus.SessionBus().get_object(MakeProgressEngine.ComposeBusName(name), "/")
		except:
			object = False
		while object:
			i += 1
			name = basename + str(i)
			try:
				object = dbus.SessionBus().get_object(MakeProgressEngine.ComposeBusName(name), "/")
			except:
				object = False
			# infinite loop check
			if i > 1000:
				break
		return name

	@staticmethod
	def ComposeBusName(name):
		return bus_name + '.make-progress.' + name



def CreateDataEngine(parent):
	return MakeProgressEngine(parent)
